CIRRO_AUTH_CLIENT_ID = 'CIRRO_AUTH_CLIENT_ID'
CIRRO_DB_URI = 'CIRRO_DB_URI'
CIRRO_DATABASE = 'CIRRO_DATABASE'
CIRRO_EMAIL = 'CIRRO_EMAIL'
CIRRO_SERVE = 'CIRRO_SERVE'
CIRRO_MAX_WORKERS = 'CIRRO_MAX_WORKERS'
CIRRO_FOOTER = 'CIRRO_FOOTER'
CIRRO_BRAND = 'CIRRO_BRAND'
CIRRO_UPLOAD = 'CIRRO_UPLOAD'
