import {format} from 'd3-format';

export const numberFormat = format('.1f');
export const numberFormat2f = format('.2f');
export const intFormat = format(',');
export const numberFormat0 = format('.0f');
